/*
  PV CPConv.
  (c) Alexey Makhmutov (aka Lerm) 2004

  Published under GPL license
*/

#include <stdrom.h>
#include <libc.h>
#include <ZiFace.h>
#include <Dialogs.h>
#include <cpconv.h>

/* Category selector */
char category[5][15];
char* catListData[]={category[0],category[1],category[2],category[3],category[4]};

Label categoryLabel={"Select memo",IB_PFONT2,ALIGN_CENTER};
ListBox catList={elements(catListData),0,0,catListData};
ComboBox categoryBox={0, 53, &catList};

/* Main list */
char* getMemoName(int);
ListBox memoList={2,0,0,NULL,(ListData)getMemoName};

/* Codepage selectors */
char* cpNames[CP_COUNT]={
                          "PV-RUS",
                          "PV-EURO",
                          "CP-1251 (Win-Ru)",
                          "CP-1252 (Euro)",
                          "CP-866  (Dos-Ru)",
                          "KOI8-R"
                        };

ListBox cpSrcList={CP_COUNT, 0, CP_1251, cpNames};
ListBox cpDstList={CP_COUNT, 0, CP_PVRUS,cpNames};

ComboBox cpSrcCombo={ 0, 53, &cpSrcList};
ComboBox cpDstCombo={ 0, 53, &cpDstList};

Label cpSrcLabel={"From:",IB_PFONT2,ALIGN_RIGHT};
Label cpDstLabel={"To:",IB_PFONT2,ALIGN_RIGHT};

/* Carriage return selector */
char* crNames[]={
                  "Keep current",
                  "CR (PV, Mac)",
                  "CR LF (DOS, Win)",
                  "LF (Unix)"
                };
ListBox crList={elements(crNames), 0, 0, crNames};
ComboBox crCombo={ 0, 43, &crList};
Label crLabel={"Carriage:",IB_PFONT2,ALIGN_RIGHT};

/* Buttons */
Button startButton={B_PRESS, 0, "Start"};
Button deleteButton={B_PRESS, 0, "Delete"};
Button aboutButton={B_PRESS, 0, "About"};
Button exitButton={B_PRESS, 0, "Exit"};

Button saveCheck={B_CHECK, 0, "Save As New"};

/* About section */
Icon aboutIcon={107};
Label aboutLabel={"CPConv for PV\rby Alexey Makhmutov\r(aka Lerm)\r\xdc 2004",IB_PFONT1,ALIGN_CENTER};

/* Handlers */
int onCategoryChange(Event*,Control*);
int onCpChange(Event*,Control*);
int onStart(Event*,Control*);
int onDelete(Event*,Control*);
int onAbout(Event*,Control*);
int onExit(Event*,Control*);

Control cList[]={
/* 00 */         {C_COMBOBOX,  80,   0,  80, 12, TRUE, TRUE, (CHandler)onCategoryChange, &categoryBox},
/* 01 */         {C_LABEL,      0,   2,  80, 12, TRUE, TRUE, NULL, &categoryLabel},
/* 02 */         {C_LIST,       0,  13, 160, 63, TRUE, TRUE, (CHandler)onCpChange, &memoList},
/* 03 */         {C_COMBOBOX,  60,  78, 100, 12, TRUE, TRUE, (CHandler)onCpChange, &cpSrcCombo},
/* 04 */         {C_COMBOBOX,  60,  92, 100, 12, TRUE, TRUE, (CHandler)onCpChange, &cpDstCombo},
/* 05 */         {C_COMBOBOX,  60, 106, 100, 12, TRUE, TRUE, (CHandler)onCpChange, &crCombo},
/* 06 */         {C_LABEL,      0,  80,  55, 12, TRUE, TRUE, NULL, &cpSrcLabel},
/* 07 */         {C_LABEL,      0,  94,  55, 12, TRUE, TRUE, NULL, &cpDstLabel},
/* 08 */         {C_LABEL,      0, 108,  55, 12, TRUE, TRUE, NULL, &crLabel},
/* 09 */         {C_BUTTON,    60, 122,  55, 12, TRUE, TRUE, NULL, &saveCheck},
/* 10 */         {C_BUTTON,   159-38*4-6, 159-15,  38, 15, TRUE, FALSE, (CHandler)onStart,  &startButton},
/* 11 */         {C_BUTTON,   159-38*3-4, 159-15,  38, 15, TRUE, FALSE, (CHandler)onDelete, &deleteButton},
/* 12 */         {C_BUTTON,   159-38*2-2, 159-15,  38, 15, TRUE, TRUE,  (CHandler)onAbout,  &aboutButton},
/* 13 */         {C_BUTTON,   159-38,     159-15,  38, 15, TRUE, TRUE,  (CHandler)onExit,   &exitButton}
                };

void formHandler(Event*,Form*);

Form mainForm={0,0,160,160,FORM_INACTIVE,elements(cList),cList,(FHandler)formHandler};

/* File operations */

FILE_BUF fb;
FILE_INFO fi;
int currentRecord;

unsigned int  far unicodeBuffer[8192];
unsigned char far symbolsBuffer[4096];
unsigned int unicodeSize;
unsigned int symbolsSize;


void updateControl(Control* control, int enabled)
{
  if (control->enabled!=enabled)
  {
    control->enabled=enabled;
    drawControl(control);
  }
}

void updateButtons()
{
  updateControl(cList+10,(cpSrcList.selected!=cpDstList.selected||crList.selected!=CR_KEEP) && memoList.selected!=-1);
  updateControl(cList+11,memoList.selected!=-1);
}

void initCategory()
{
  int i=0;

  for(;i<5; i++)
  {
    LibLblAreaRead(category[i],5+i);
    if (*category[i]==0)
    {
      LibReadMessage(246+i,category[i]);
    }
  }
  catList.selected=SYS_MEMO_CATEGORY==0?0:SYS_MEMO_CATEGORY-1;
}

void readCategory()
{
  fb.fsb_main_entry_=0x30; /* Memo */ 
  fb.fsb_sub_entry_=catList.selected+1;
  fb.fsb_scrt_info_=FILE_OPEN_MODE;

  fi.kind=FILE_KIND_TEXT;

  memoList.size=LibGetFileCnt(&fb);
  memoList.first=0;
  memoList.selected=-1;
  currentRecord=-1;
}

void nextRecord()
{
  LibFileFindNext(&fb, &fi, 0x00);
}

void readRecord(int i)
{
  if (i<currentRecord||currentRecord==-1)
  {
    fi.fp=0xFFFF;
    nextRecord();
    currentRecord=0;
  }

  while(i>currentRecord)
  {
    nextRecord();
    currentRecord++;
  }

  LibFileRead(&fb, &fi);;
}

void writeRecord()
{
  LibFileWrite(&fb, &fi);
}

void prepareName()
{
  char* point=fb.fbuf.text.fsb_text_buf_;
  char c;
  for(c=*point;c!=0&&c!='\r';c=*(point++))
    ;
  *point=0;
}

char* getMemoName(int i)
{
  readRecord(i);
  prepareName();
  return fb.fbuf.text.fsb_text_buf_;
}

int onCategoryChange(Event* e, Control* c)
{
  readCategory();
  initControl(cList+2);
  drawControl(cList+2);
  updateButtons();
  SYS_MEMO_CATEGORY=catList.selected+1;
  return 1;
}

int onCpChange(Event* e, Control* c)
{
  updateButtons();
  return 1;
}

int encodeCharacter(int index, unsigned int far *dstCP, int transType)
{
  int unicode=unicodeBuffer[index];
  int symbol=0;
  
  if (unicode<128)
  {
    if (transType!=CR_KEEP)
    {
      int currentType=CR_KEEP;

      /* Detect current CR */
      if (unicode==CODE_LF)
      {
        currentType=CR_LF;
      }
      else if (unicode==CODE_CR)
      {
        if (index==unicodeSize-1||unicodeBuffer[index+1]!=CODE_LF)
        {
          currentType=CR_CR;
        }
        else
        {
          currentType=CR_CRLF;
        }
      }

      /* Translate only carriage returns */
      if (currentType!=CR_KEEP)
      {
        if (transType==CR_CR)
        {
          symbol=CODE_CR;
        }
        else 
        {
          if (transType==CR_CRLF)
          {
            symbolsBuffer[symbolsSize++]=CODE_CR;
          }
          symbol=CODE_LF;
        }

        if (currentType==CR_CRLF)
        {
          index++;
        }
      }
    }

    if (symbol==0)
    {
      symbol=(unsigned char)unicode;
    }
  }
  else
  {
    for(symbol=0; symbol<128&&dstCP[symbol]!=unicode; symbol++)
       ;

    if (symbol==128)
    {
      //TODO: borders
      symbol=0x20;
    }
    else
    {
      symbol+=128;
    }
  }
  symbolsBuffer[symbolsSize++]=symbol;

  return ++index;
}

void encodeRecord()
{
  unsigned int far *srcCP=unicodeTable[cpSrcList.selected];
  unsigned int far *dstCP=unicodeTable[cpDstList.selected];
  int transType=crList.selected;
  int i;

  for(unicodeSize=0; fb.fbuf.text.fsb_text_buf_[unicodeSize]!=0; unicodeSize++)
  {
    unsigned char c=fb.fbuf.text.fsb_text_buf_[unicodeSize];

    unicodeBuffer[unicodeSize]=c<128?c:srcCP[c-128];
  }

  for(i=0,symbolsSize=0; i<unicodeSize; i=encodeCharacter(i,dstCP,transType))
    ;
}

int onStart(Event* e, Control* c)
{
  if (dialogSelect(DICON_HAND, "Encode?"))
  {
    showWaitWindow(DICON_WAIT);
    readRecord(memoList.selected);
    encodeRecord();
    /* Now we have encoded record in symbols buffer */
    if (symbolsSize==2049&&symbolsBuffer[2047]==CODE_CR&&symbolsBuffer[2048]==CODE_LF)
    { 
      symbolsSize-=2;
    }

    if (symbolsSize>2048)
    {/* Save as two new records */
      int i,j;

      if (symbolsBuffer[2047]==CODE_CR&&symbolsBuffer[2048]==CODE_LF)
      {
        symbolsBuffer[2047]=0;
        symbolsBuffer[2048]=0x20;
      }
      dialogInfo(DICON_EXCL,"Message will be\rsplitted!");
      for(i=2048,j=0; i<symbolsSize; i++,j++)
      {
        fb.fbuf.text.fsb_text_buf_[j]=symbolsBuffer[i];
      }
      fb.fbuf.text.fsb_text_buf_[j]=0;
      fi.fp=0xFFFF;
      writeRecord();
      for(i=0; i<2048; i++)
      {
        fb.fbuf.text.fsb_text_buf_[i]=symbolsBuffer[i];
      }
      fb.fbuf.text.fsb_text_buf_[i]=0;
      fi.fp=0xFFFF;
      writeRecord();
      memoList.selected=0;
      memoList.first=0;
    }
    else
    {
      int i;
      for(i=0; i<symbolsSize; i++)
      {
        fb.fbuf.text.fsb_text_buf_[i]=symbolsBuffer[i];
      }
      fb.fbuf.text.fsb_text_buf_[i]=0;

      if (saveCheck.state)
      {/* Save as new */
        fi.fp=0xFFFF;
        memoList.selected=0;
        memoList.first=0;
      }
      writeRecord();
    }
    hideWaitWindow();
    /* Redraw window after encoding */
    currentRecord=-1;
    memoList.size=LibGetFileCnt(&fb);
    drawControl(cList+2);
  }
  return 1;
}

int onDelete(Event* e, Control* c)
{
  if (dialogSelect(DICON_HAND,"Delete this record?"))
  {
    showWaitWindow(DICON_TRASH);
    readRecord(memoList.selected);
    LibFileRemove(&fb, &fi);
    hideWaitWindow();

    memoList.size=LibGetFileCnt(&fb);

    if (memoList.selected>=memoList.size)
    {
      memoList.selected--;
    }

    showListElement(&memoList, memoList.selected, cList[2].h);
    currentRecord=-1;
    drawControl(cList+2);

    updateButtons();
  }
  return 1;
}

int onAbout(Event* e, Control* c)
{
  dialogInfo(DICON_NONE,"\xDC Alexey Makhmutov\r(aka Lerm)\r2004");
  return 1;
}

int onExit(Event* e, Control* c)
{
  mainForm.status=FORM_INACTIVE;
  return 1;
}

void formHandler(Event* e, Form* f)
{
  if (e->type==E_LEVER)
  {/* Change category with left-right */
    int lever=e->lever;
    int newCat=catList.selected;
    if (lever==LPSW_LEFT)
    {
      newCat=newCat==0?4:newCat-1;
    }
    else if (lever==LPSW_RIGHT)
    {
      newCat=newCat==4?0:newCat+1;
    }

    if (newCat!=catList.selected)
    {
      catList.selected=newCat;
      drawControl(cList);
      onCategoryChange(e,cList);
    }
  }
  else if (e->type==E_ESC)
  {
    f->status=FORM_INACTIVE;
  }
}

void main()
{ 
  initCategory();
  readCategory();
  
  LibTchInit();
  LibTchStackClr();

  initForm(&mainForm);
  processForm(&mainForm);

  LibJumpMenu();
}
