/*
  ZProg PV Interface.
  (c) Alexey Makhmutov (aka Lerm) 2003,2004

  Published under GPL license
*/

#include <define.h>
#include <stdrom.h>
#include <libc.h>
#include <dialogs.h>
#include <ZiFace.h>

/*
  Borders:
  107 - 128x50
  108 - 128x62
  109 - 128x76
*/
static Icon borderIcon={108};
static Icon signIcon  ={46};

static Label messageLabel={NULL,IB_PFONT1,ALIGN_CENTER};
static Button bOk={B_PRESS,FALSE,"Ok"};
static Button bYes={B_PRESS,FALSE,"Yes"};
static Button bNo={B_PRESS,FALSE,"No"};

#define DIALOG_BASE 49

int onOk(Event*,Control*);
int onYes(Event*,Control*);
int onNo(Event*,Control*);
void onForm(Event*,Form*);

static Control dialog1[]={
                  {C_ICON,   16, DIALOG_BASE,    128, 62, TRUE, TRUE, NULL, &borderIcon},
                  {C_ICON,   24, DIALOG_BASE+11,  24, 24, TRUE, TRUE, NULL, &signIcon},
                  {C_LABEL,  50, DIALOG_BASE+15,  90,  8, TRUE, TRUE, NULL, &messageLabel},
                  {C_BUTTON, 63, DIALOG_BASE+41,  40, 15, TRUE, TRUE, (CHandler)onOk, &bOk}
                 };
static Form dialogForm1={16,DIALOG_BASE,128,62,FORM_INACTIVE,elements(dialog1),dialog1,(FHandler)onForm};

static Control dialog2[]={
                  {C_ICON,   16, DIALOG_BASE,    128, 62, TRUE, TRUE, NULL, &borderIcon},
                  {C_ICON,   24, DIALOG_BASE+11,  24, 24, TRUE, TRUE, NULL, &signIcon},
                  {C_LABEL,  50, DIALOG_BASE+15,  90,  8, TRUE, TRUE, NULL, &messageLabel},
                  {C_BUTTON, 56, DIALOG_BASE+41,  40, 15, TRUE, TRUE, (CHandler)onYes, &bYes},
                  {C_BUTTON, 98, DIALOG_BASE+41,  40, 15, TRUE, TRUE, (CHandler)onNo, &bNo}
                 };
static Form dialogForm2={16,DIALOG_BASE,128,62,FORM_INACTIVE,elements(dialog2),dialog2,(FHandler)onForm};

static byte far iconsType[]={10,11,24,25,28,46,55};
static int selectResult;

static int onOk(Event* e, Control* c)
{
  if (e->type==E_PEN)
  {
    dialogForm1.status=FORM_INACTIVE;
    return 1;
  }
  return 0;
}

static int onSelect(Event* e, Control* c, int res)
{
  if (e->type==E_PEN)
  {
    dialogForm2.status=FORM_INACTIVE;
    selectResult=res;
    return 1;
  }
  return 0;
}

static int onYes(Event* e, Control* c)
{
  return onSelect(e,c,1);
}

static int onNo(Event* e, Control* c)
{
  return onSelect(e,c,0);
}

static void onForm(Event* e, Form* f)
{
  if (e->type==E_ESC)
  {
    f->status=FORM_INACTIVE;
    selectResult=0;
  }
}

static void executeDialog(Form* form, byte iconType, char* message)
{
  Control* cList=form->list;
  if (iconType==DICON_NONE)
  {
    cList[1].visible=FALSE;
    cList[2].x=20;
    cList[2].w=120;
    cList[2].y=DIALOG_BASE+10;
  }
  else
  {
    cList[1].visible=TRUE;
    cList[2].x=50;
    cList[2].w=90;
    cList[2].y=DIALOG_BASE+15;
    signIcon.no=iconsType[iconType-1];
  }
  messageLabel.label=message;

  initForm(form);
  processForm(form);
}

void dialogInfo(byte iconType, char* message)
{
  executeDialog(&dialogForm1,iconType,message);
}

int dialogSelect(byte iconType, char* message)
{
  executeDialog(&dialogForm2,iconType,message);
  return selectResult;
}

void showWaitWindow(int icon)
{
  LibOpenWindow(16,55,128,50);
  LibPutFarData(16,55,107);
  if (icon!=DICON_NONE)
  {
    LibPutFarData(68,68,iconsType[icon-1]);
  }
  LibPutDispBox(16,55,128,50);
}

void hideWaitWindow()
{
  LibCloseWindow();
  LibPutDispBox(16,55,128,50);
}