/*
  ZProg PV Interface.
  (c) Alexey Makhmutov (aka Lerm) 2003,2004
*/

#ifndef LERM_ZIFACE_H_
#define LERM_ZIFACE_H_

#include <define.h>

#ifndef elements
#define elements(list) sizeof(list)/sizeof(*list)
#endif

#define getControl(x,y) ((y*)(x->data))

/* Form statuses */
#define FORM_INACTIVE 0
#define FORM_ACTIVE   1

/* Types of events */
#define E_ALL      0
#define E_PEN      1
#define E_LEVER    2
#define E_TIMER    3
#define E_ESC      4
#define E_MENU     5
#define E_ACTIVATE 6
#define E_COMPLETE 7
#define E_CANCEL   8

/* You can create your own events with this type or greater */
#define E_CUSTOM   32

/* Lever type */
#define LPSW_NONE  0
#define LPSW_PUSH  1
#define LPSW_UP    2
#define LPSW_DOWN  4
#define LPSW_LEFT  8
#define LPSW_RIGHT 16

/* Types of controls */
#define C_LABEL    0
#define C_ICON     1
#define C_BITMAP   2
#define C_BUTTON   3
#define C_RGROUP   4
#define C_LIST     5
#define C_EDIT     6
#define C_PROGRESS 7
#define C_COUNTBOX 8
#define C_COMBOBOX 9
#define C_BORDER   10

/* SubTypes of buttons */
#define B_PRESS 0
#define B_CHECK 1
#define B_RADIO 2

/* Label align */
#define ALIGN_LEFT   0
#define ALIGN_RIGHT  1
#define ALIGN_CENTER 2

/* Define structures */
/* Main objects */
typedef struct Event Event;
typedef struct Control Control;
typedef struct Form Form;

/* Controls */
typedef struct Label Label;
typedef struct Icon Icon;
typedef struct Bitmap Bitmap;
typedef struct Button Button;
typedef struct RadioGroup RadioGroup;
typedef struct ListBox ListBox;
typedef struct EditBox EditBox;
typedef struct ProgressBar ProgressBar;
typedef struct CountBox CountBox;
typedef struct ComboBox ComboBox;

/* General event handler */
typedef int (*CHandler)(Event*,Control*);
typedef void (*FHandler)(Event*,Form*);

typedef char* (*ListData)(int);

struct Event
{
  int type;    /* Type of event: see define above */
  int x,y;     /* Touch coords */
  int lever;   /* State of lever: see define below */
};

struct Control
{
  int type;          /* Control type: see define above */

  int x,y,w,h;       /* Control coord */
  byte visible;      /* Visible state */
  byte enabled;      /* Enabled state */
  CHandler custHand; /* Custom handler */
  void* data;        /* Control type specific data */
};

struct Form
{
  int x,y,w,h;    /* Form coords    */
  int status;     /* Form status: FORM_INACTIVE, FORM_ACTIVE */
  word size;      /* Controls count */
  Control* list;  /* Controls list  */
  FHandler hand;  /* Form handler   */
};

struct Label
{
  char* label;  /* Label text */
  byte  font;   /* Label font: IB_PFONT1, IB_PFONT2, IB_PFONT3, IB_CG57FONT */
  byte  align;  /* Align: ALIGN_LEFT, ALIGN_RIGHT, ALIGN_CENTER */
};

struct Icon
{
  int no;       /* Icon number */
};

struct Bitmap
{
  byte far *image;  /* Pointer to image in PV format */
};

struct Button
{
  int subType;   /* Subtype of button: B_PRESS, B_CHECK, B_RADIO */

  int state;     /* State of button */
  char* label;   /* Button text */
  int font;      /* Font type (optional) */
};

struct RadioGroup
{
  int size;       /* Buttons count */
  Button** group; /* Array of button pointers */
};

struct ListBox
{
  int size;       /* Items count */
  int first;      /* First visible item */
  int selected;   /* Selected item */
  char** list;    /* Array of string pointers */
  ListData data;  /* Data provider (optional) */
};

struct EditBox
{
  int maxNum;     /* Maximum number of symbols */
  char* buffer;   /* Pointer to buffer */
};

struct ProgressBar
{
  unsigned long size;    /* Total values  */
  unsigned long current; /* Current value */
};

struct CountBox
{
  int current; /* Current value */
  int minimum; /* Minimum value */
  int maximum; /* Maximum value */
  int step;    /* Step          */
};

struct ComboBox
{
  int state;         /* 0 - Closed, 1 - Open */ /* TODO: make constant */
  int height;        /* Height of open list */
  ListBox *listBox;  /* Base list */
};

/* Elements function */
int showListElement(ListBox* list, int number, int listHeight);

/* Core functions prototypes */

void drawControl(Control* control);
 /* Draw control */

void initControl(Control* control);
 /* Init control */
 
void initForm(Form* form);
 /* Init all controls in form */

void drawForm(Form* form);
 /* Draw entire form */

void processForm(Form* form);
 /* Start event dispatching for form */

int createEvent(int type, int x, int y, int lever);
 /* Put event into queue */

int createHook(int type, Control* control);
 /* Hook event type */

int removeHook(int type, Control* control);
 /* Remove hook */

#endif
